/*
 * Copyright © 2016 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <dlfcn.h>
#include <unistd.h>
#include <stdlib.h>

#include <string>

namespace
{
bool file_exists(const char *path)
{
    return access(path, F_OK ) == 0;
}

bool is_absolute_path(std::string const& path)
{
    return !path.empty() && path.at(0) == '/';
}

std::string snapify(const char *path)
{
    if (file_exists(path))
        return {path};

    std::string the_path{path};
    if (is_absolute_path(the_path))
    {
        std::string snap_path{getenv("SNAP")};
        the_path.insert(0, snap_path);
        return the_path;
    }
    else
    {
        return {path};
    }
}
}

extern "C"
{
typedef void *(*orig_dlopen_f_type)(const char *pathname, int flags);
typedef int(*orig_execvp_f_type)(const char *file, char *const argv[]);
typedef int(*orig_execve_f_type)(const char *file, char *const argv[], char *const envp[]);

/**
 * These shims prepend the snap path if they detect access to non-existent
 * absolute path
 */

void *dlopen(const char *filename, int flags)
{
    orig_dlopen_f_type orig_dlopen;
    orig_dlopen = (orig_dlopen_f_type)dlsym(RTLD_NEXT, "dlopen");

    if (filename == nullptr)
    {
        return orig_dlopen(filename, flags);
    }

    auto lib_name = snapify(filename);
    return orig_dlopen(lib_name.c_str(), flags);
}

int execvp(const char *file, char *const argv[])
{
    orig_execvp_f_type orig_execvp;
    orig_execvp = (orig_execvp_f_type)dlsym(RTLD_NEXT, "execvp");

    auto file_name = snapify(file);
    return orig_execvp(file_name.c_str(), argv);
}

int execve(const char *file, char *const argv[], char *const envp[])
{
    orig_execve_f_type orig_execve;
    orig_execve = (orig_execve_f_type)dlsym(RTLD_NEXT, "execve");

    auto file_name = snapify(file);
    return orig_execve(file_name.c_str(), argv, envp);
}
}
